#include "cfs.h"
#include "types.h"

const int min_granularity = 2;
const int latency = (NPROC >> 1);

struct RBTree queue;
struct RBTree *rdy_queue = &queue;

void
rotateLeft(struct RBTree* tree, struct proc* ptr){
  struct proc *parent = ptr->rb_parent;
  struct proc *right_ = ptr->right;
  ptr->right = right_->left;
  if(right_->left != 0){
    right_->left->rb_parent = ptr;
  }
  ptr->rb_parent = right_;
  right_->left = ptr;
  if(parent == 0){
    tree->root = right_;
  } else {
    if(parent->left == ptr){
      parent->left = right_;
    } else {
      parent->right = right_;
    }
  }
  right_->rb_parent = parent;
}

void 
rotateRight(struct RBTree* tree, struct proc* ptr){
  struct proc *left_ = ptr->left;
  struct proc *parent = ptr->rb_parent;
  ptr->left = left_->right;
  if(left_->right != 0){
    left_->right->rb_parent = ptr;
  }
  left_->right = ptr;
  ptr->rb_parent = left_;
  if(parent == 0){
    tree->root = left_;
  } else {
    if(parent->left == ptr){
      parent->left = left_;
    } else {
      parent->right = left_;
    }
  }
  left_->rb_parent = parent;
}

struct proc*
getGrandparent(struct proc* process){
  if(process != 0 && process->rb_parent != 0){
	  return process->rb_parent->rb_parent;
  } 
  return 0;
}

struct proc*
getUncle(struct proc* process){
  struct proc* grandParent = getGrandparent(process);
  if(grandParent != 0){
    if(process->rb_parent == grandParent->left){
      return grandParent->right;
    } else {
      return grandParent->left;
    }
  }
  return 0;
}

int
insertFix(struct RBTree *tree, struct proc *node){
  if(node->rb_parent == 0){
    node->color = BLACK;
    return 0;
  }
  struct proc *parent = node->rb_parent;
  struct proc *grandparent = getGrandparent(node);
  if(grandparent != 0){
    if(parent->color == RED){
      struct proc *uncle = getUncle(node);
      if(uncle != 0 && uncle->color == RED){
        parent->color = BLACK;
        uncle->color = BLACK;
        grandparent->color = RED;
        insertFix(tree, grandparent);
      } else {
        if(parent == grandparent->left) {
          if(parent->right == node){
            rotateLeft(tree, parent);
          }
          parent->color = BLACK;
          grandparent->color = RED;
          rotateRight(tree, grandparent);
          return 0;
        } else {
          if(parent->left == node) {
            rotateRight(tree, parent);
          }
          parent->color = BLACK;
          grandparent->color = RED;
          rotateLeft(tree, grandparent);
          return 0;
        }
      }
    }
  } else {
    if(parent->color == RED) {
      parent->color = BLACK;
    }
  }
  return 0;
}

int
insertProc(struct RBTree *tree, struct proc *node){
  node->rb_parent = 0;
  node->left = 0;
  node->right = 0;
  node->next = 0;


  acquire(&tree->lock);

  node->weight = getWeight(node->nice);
  tree->rbTreeWeight += node->weight;

  struct proc *ptr = tree->root;
  node->color = RED;
  if(ptr == 0){
    tree->root = node;
  } else {
    while(1){
      if(ptr->vruntime == node->vruntime) {
        break;
      }
      if(ptr->vruntime < node->vruntime){
        if(ptr->right != 0)
          ptr = ptr->right;
        else
          break;
      } else {
        if(ptr->left != 0)
          ptr = ptr->left;
        else
          break;
      }
    }
    if(ptr->vruntime < node->vruntime){
      ptr->right = node;
      node->rb_parent = ptr;
      insertFix(tree, node);
    } else if(ptr->vruntime > node->vruntime) {
      ptr->left = node;
      node->rb_parent = ptr;
      insertFix(tree, node);
    } else {
      node->next = ptr->next;
      ptr->next = node;
    }
  }
  if(tree->min_vruntime_proc == 0 || tree->min_vruntime_proc->left != 0)
    tree->min_vruntime_proc = min_vruntime_proc(tree->root);
  tree->count++;
  // printf("enqueue: %d\n", pid);
  release(&tree->lock);
  return 0;
}

void printTreeHelper(struct proc *node, int space) {
    if (node == 0) {
        return;
    }
    space += 10;
    printTreeHelper(node->right, space);
    printf("\n");
    for (int i = 10; i < space; i++) {
        printf(" ");
    }
    printTreeHelper(node->left, space);
}

void printRBTree() {
    printf("Red-Black Tree:\n");
    printTreeHelper(rdy_queue->root, 0);
}

struct proc*
min_vruntime_proc(struct proc* node){
  if(node != 0){
    if(node->left != 0){
      return min_vruntime_proc(node->left);
    } else {
      return node;
    }
  }
	return 0;
}

int
remove_fix(struct RBTree *tree, struct proc *node, struct proc *parent) {
  if((node != 0 && node->color == RED) || parent == 0) {
    node->color = BLACK;
    return 0;
  }
  struct proc *sibling;
  struct proc *sib_left;
  struct proc *sib_right;
  if(parent->left == node) {
    sibling = parent->right;
    //sibling cannot be null, since the left wing has black height 1
    sib_left = sibling->left;
    sib_right = sibling->right;
    if(sibling->color == RED) {
      rotateLeft(tree, parent);
      parent->color = RED;
      sibling->color = BLACK;
      remove_fix(tree, node, parent);
    } else {
      if((sib_left == 0 || sib_left->color == BLACK) && (sib_right == 0 || sib_right->color == BLACK)) {
        sibling->color = RED;
        if(parent->color == RED) {
          parent->color = BLACK;
        } else {
          remove_fix(tree, parent, parent->rb_parent);
        }
      } else if(sib_right == 0 || sib_right->color == BLACK) 
      {
        sibling->color = RED;
        sib_left->color = BLACK;
        rotateRight(tree, sibling);
        remove_fix(tree, node, parent);
      } else {
        rotateLeft(tree, parent);
        sibling->color = parent->color;
        parent->color = BLACK;
        sib_right->color = BLACK;
        if(sibling == tree->root) {
          sibling->color = BLACK;
        }
      }
    }
  } else {
    sibling = parent->left;
    //sibling cannot be null, since the left wing has black height 1
    sib_left = sibling->left;
    sib_right = sibling->right;
    if(sibling->color == RED) {
      rotateRight(tree, parent);
      parent->color = RED;
      sibling->color = BLACK;
      remove_fix(tree, node, parent);
    } else {
      if((sib_left == 0 || sib_left->color == BLACK) && (sib_right == 0 || sib_right->color == BLACK)) {
        sibling->color = RED;
        if(parent->color == RED) {
          parent->color = BLACK;
        } else {
          remove_fix(tree, parent, parent->rb_parent);
        }
      } else if(sib_left == 0 || sib_left->color == BLACK) 
      {
        sibling->color = RED;
        sib_right->color = BLACK;
        rotateLeft(tree, sibling);
        remove_fix(tree, node, parent);
      } else {
        rotateRight(tree, parent);
        sibling->color = parent->color;
        parent->color = BLACK;
        sib_left->color = BLACK;
        if(sibling == tree->root) {
          sibling->color = BLACK;
        }
      }
    }
  }

  return 0;
}

struct proc*
rdyqueue_deq(struct RBTree *tree) {
  acquire(&tree->lock);
  if(tree->count == 0){
    release(&tree->lock);
    return 0;
  }
  struct proc *leftmost = min_vruntime_proc(tree->root);
  if(leftmost->state != RUNNABLE)
    panic("not runnable");
  
  //delete the leftmost node
  struct proc *right_ = leftmost->right;
  struct proc *parent = leftmost->rb_parent;
  
  int color = leftmost->color;

  if(leftmost->next == 0){

    if (parent != 0) {
      parent->left = right_;
    } else {
        tree->root = right_;
    }


    if (right_ != 0) {
      right_->rb_parent = parent;
    }
   

    if (color == BLACK && tree->root != 0) {
      remove_fix(tree, right_, parent);
    }
  } else {
    struct proc* list_next = leftmost->next;
    leftmost->next = list_next->next;
    list_next->next = 0;
    leftmost = list_next; //replace back
  }

  

  tree->min_vruntime_proc = min_vruntime_proc(tree->root);
  leftmost->timeSlice = (tree->period * leftmost->weight / tree->rbTreeWeight);
	tree->rbTreeWeight -= leftmost->weight;
  if(tree->count > (latency / min_granularity)){
		tree->period = tree->count * min_granularity;//更新调度周期
	}


  tree->count--;
  // printf("deq proc name: %s\n", leftmost->name);
  release(&tree->lock);
  leftmost->next = 0;
  leftmost->left = 0;
  leftmost->right = 0;
  leftmost->rb_parent = 0;
  return leftmost;
}

void
rbinit(struct RBTree *tree)
{
  initlock(&tree->lock, "RBTree");
  tree->count = 0;
  tree->root = 0;
  tree->rbTreeWeight = 0;
  tree->min_vruntime_proc = 0;
  tree->period = latency; //initial value for sheduling period
}

int preemptionTest(struct proc* current, struct proc* min_vruntime) {
  // printf("readyQueue count: %d\n", rdy_queue->count);
  int time_used = current->curruntime;
  // printf("current->curruntime: %d, curname: %s\n", time_used, current->name);
  if(time_used >= min_granularity) {
      if ((time_used >= current->timeSlice)) {
        return 1;
      }
      if (min_vruntime != 0  && current->vruntime > min_vruntime->vruntime) {
          if ((time_used != 0) && (time_used >= min_granularity)) {
              return 1;
          } else if (time_used == 0) {
              return 1;
          }
      }
  }
  return 0;
}



const int nice_weights[40] = {
/* -20 */     88761,     71755,     56483,     46273,     36291,
/* -15 */     29154,     23254,     18705,     14949,     11916,
/* -10 */      9548,      7620,      6100,      4904,      3906,
/*  -5 */      3121,      2501,      1991,      1586,      1277,
/*   0 */      1024,       820,       655,       526,       423,
/*   5 */       335,       272,       215,       172,       137,
/*  10 */       110,        87,        70,        56,        45,
/*  15 */        36,        29,        23,        18,        15,
};

int getWeight(int nice) {
    if (nice < MIN_NICE) {
        nice = MIN_NICE;
    } else if (nice > MAX_NICE) {
        nice = MAX_NICE;
    }
    return nice_weights[nice - MIN_NICE];
}