#include "buddy2.h"
#include "param.h"
#include "include/spinlock.h"
#include "sleeplock.h"
#include "riscv.h"
#include "defs.h"
#include "fs.h"
#include "buf.h"
#include "slub.h"
enum {UNUSED, ALLO, SPLIT, INLIST};

struct buddy_ bdsys;

void buddy_init_(uint64 base){
    bdsys.base_addr = PGROUNDUP(base);
    for(int i = 1; i < (1 << (LAYER_NUM + 1)) - 1; i++){
        bdsys.nodes[i].next = 0;
        bdsys.nodes[i].prev = 0;
        bdsys.nodes[i].state = UNUSED;
        bdsys.nodes[i].id = i;
        
    }
    for(int i = 0; i < LAYER_NUM; i++)
        bdsys.freelists[i] = 0;
    bdsys.freelists[0] = &bdsys.nodes[0];
    bdsys.nodes[0].state = INLIST;
    bdsys.nodes[0].id = 0;
    bdsys.nodes[0].prev = 0;
    bdsys.nodes[0].next = 0; 
}

//add the node to the head of the linked list
int add_list(uint32 level, struct tree_node *ptr){
    if(ptr != 0 && level < LAYER_NUM){
        if(bdsys.freelists[level] != 0){
            ptr->next = bdsys.freelists[level];
            ptr->next->prev = ptr;
        } else {
            ptr->next = 0;
        }
        ptr->prev = 0;//not quite sure
        bdsys.freelists[level] = ptr;
        return 0;
    }
    return -1;
}

int delete_list(uint32 level, struct tree_node *ptr){
    if(ptr == 0 || level >= LAYER_NUM)
        return -1;
    if(ptr->prev == 0) {
        bdsys.freelists[level] = ptr->next;
        if(ptr->next != 0){
            ptr->next->prev = 0;
        }
    } else {
        ptr->prev->next = ptr->next;
        if(ptr->next != 0){
            ptr->next->prev = ptr->prev;
        }
    }
    ptr->prev = 0;
    ptr->next = 0;
    return 0;
}

//size : num of pages
uint64 buddy_alloc_(uint64 size){
    int i = 1;
    int final_id;
    //lock acquire
    while((1 << (i - 1)) < size || bdsys.freelists[LAYER_NUM - i] == 0){
        if(i == LAYER_NUM){
            return 0;
        }
        i++;
    }
    if((i == 1 || ((1 << (i - 2))) < size)){ //no need to split
        bdsys.freelists[LAYER_NUM - i]->state = ALLO;
        final_id = bdsys.freelists[LAYER_NUM - i]->id;
        delete_list(LAYER_NUM - i, bdsys.freelists[LAYER_NUM - i]);
    } else {
        while(i > 1 && (1 << (i - 2)) >= size) {
            bdsys.freelists[LAYER_NUM - i]->state = SPLIT;
            uint32 id = bdsys.freelists[LAYER_NUM - i]->id;
            delete_list(LAYER_NUM - i, bdsys.freelists[LAYER_NUM - i]);
            i--;
            struct tree_node *right_ch = &bdsys.nodes[RIGHT(id)];
            right_ch->state = INLIST;
            struct tree_node *left_ch = &bdsys.nodes[LEFT(id)];
            left_ch->state = INLIST;
            add_list(LAYER_NUM - i, right_ch);
            add_list(LAYER_NUM - i, left_ch);
        }
        bdsys.freelists[LAYER_NUM - i]->state = ALLO;
        final_id = bdsys.freelists[LAYER_NUM - i]->id;
        delete_list(LAYER_NUM - i, bdsys.freelists[LAYER_NUM - i]);
    }
    int offset_in_level = final_id - (1 << (LAYER_NUM - i)) + 1;
    
    uint64 result = bdsys.base_addr + (uint64)((offset_in_level * (1 << (i - 1))) << 12);
    int start = PA2PGREF_ID((uint64)result);
    
    for(int j = start; j < start + (1 << (i - 1)); j++){
        pages[j].slab = 0;
        pages[j].flags = PAGE_BUDDY;
        pages[j].header = 0;
        pages[j].freelist = 0;
        pages[j].next = 0;
    }
    return result;
}

void buddy_free_(uint64 addr){
    if((addr & 0xfff) != 0)
        panic("buddy_free_ unaligned address");
    uint32 offset = (addr - bdsys.base_addr) >> 12;
    uint32 idx = (1 << (LAYER_NUM - 1)) - 1 + offset;
    uint8 i = 1;
    while(bdsys.nodes[idx].state != ALLO){
        if(idx == 0)
            panic("buddy_free_: can't return");
        idx = PARENT(idx);
        i++;
    }

    bdsys.nodes[idx].state = INLIST; //free the node

    int start = PA2PGREF_ID((uint64)addr);
    for(int j = start; j < start + (1 << (i - 1)); j++){
        pages[j].slab = 0;
        pages[j].flags = PAGE_FREE;
        pages[j].header = 0;
        pages[j].freelist = 0;
        pages[j].next = 0;
    }
    memset((void*)addr, 0,(1 << (i - 1)));
    add_list(LAYER_NUM - i, &bdsys.nodes[idx]);
    //add list
    uint32 parent_id;
    int flag = 0;
    while(flag != 1 && idx != 0){
        parent_id = PARENT(idx);
        struct tree_node *right_ch = &bdsys.nodes[RIGHT(parent_id)];
        struct tree_node *left_ch = &bdsys.nodes[LEFT(parent_id)];
        if(left_ch->state == right_ch->state && right_ch->state == INLIST){
            delete_list(LAYER_NUM - i, left_ch);
            delete_list(LAYER_NUM - i, right_ch);
            left_ch->state = UNUSED;
            right_ch->state = UNUSED;
            add_list(LAYER_NUM - i - 1, &bdsys.nodes[parent_id]);
            bdsys.nodes[parent_id].state = INLIST;
        }
        i++;
        idx = parent_id;
    }
}

void print_free_lists() {
    printf("Free lists:\n");
    for (int level = 0; level < LAYER_NUM; level++) {
        printf("Level %d: ", level);
        struct tree_node *current = bdsys.freelists[level];

        if (current == 0) {
            printf("(empty)\n");
            continue;
        }

        while (current != 0) {
            printf("[ID: %d] ", current->id);
            current = current->next;
        }
        printf("\n");
    }
}

void test_buddy_system() {
    // Test initialization
    printf("Testing buddy_init_...\n");
    buddy_init_(0x1000);
    print_free_lists();


    // Test adding nodes to the free list
    printf("Testing add_list...\n");
    struct tree_node *node1 = &bdsys.nodes[1];
    struct tree_node *node2 = &bdsys.nodes[2];
    add_list(3, node1);
    add_list(4, node2);
    print_free_lists();

    // // Test deleting nodes from the free list
    printf("Testing delete_list...\n");
    delete_list(3, node1);
    delete_list(4, node2);
    print_free_lists();

    // Test buddy allocation
    printf("Testing buddy_alloc_...\n");
    uint64 addr = buddy_alloc_(16);
    if (addr != -1) {
        printf("Allocated block at address: 0x%p\n", (void*)addr);
    } else {
        printf("Allocation failed\n");
    }
    print_free_lists();

    // Test split and further allocations
    printf("Testing buddy_alloc_ with smaller size...\n");
    addr = buddy_alloc_(8);
    if (addr != -1) {
        printf("Allocated block at address: %p\n", (void*)addr);
    } else {
        printf("Allocation failed\n");
    }
    print_free_lists();

    // Test edge case allocation
    printf("Testing edge case allocations...\n");
    addr = buddy_alloc_(1);
    if (addr != -1) {
        printf("Allocated block at address: %p\n", (void*)addr);
    } else {
        printf("Allocation failed\n");
    }
    print_free_lists();

    addr = buddy_alloc_(1 << (BUDDY_SYSTEM_SIZE - 1));
    if (addr != -1) {
        printf("Allocated block at address: 0x%lx\n", addr);
    } else {
        printf("Allocation failed\n");
    }
    print_free_lists();
    test_buddy_free();
}

void test_buddy_free() {
    printf("Testing buddy_free_ intensively...\n");

    buddy_init_(0x1000); // Initialize the buddy system

    // Allocate various block sizes
    unsigned long long addr1 = buddy_alloc_(16);
    unsigned long long addr2 = buddy_alloc_(32);
    unsigned long long addr3 = buddy_alloc_(64);
    unsigned long long addr4 = buddy_alloc_(8);

    if (addr1 != (unsigned long long)-1) printf("Allocated block at address: 0x%llx\n", addr1);
    if (addr2 != (unsigned long long)-1) printf("Allocated block at address: 0x%llx\n", addr2);
    if (addr3 != (unsigned long long)-1) printf("Allocated block at address: 0x%llx\n", addr3);
    if (addr4 != (unsigned long long)-1) printf("Allocated block at address: 0x%llx\n", addr4);

    print_free_lists();

    // Free some blocks and check the free lists
    printf("Freeing block at address: 0x%llx\n", addr1);
    buddy_free_(addr1);
    print_free_lists();

    printf("Freeing block at address: 0x%llx\n", addr3);
    buddy_free_(addr3);
    print_free_lists();

    // Allocate again to observe fragmentation handling
    unsigned long long addr5 = buddy_alloc_(16);
    printf("Allocated block at address: 0x%llx (expected to reuse a split block)\n", addr5);
    print_free_lists();

    // Free all remaining blocks to test coalescing
    printf("Freeing block at address: 0x%llx\n", addr2);
    buddy_free_(addr2);
    print_free_lists();

    printf("Freeing block at address: 0x%llx\n", addr4);
    buddy_free_(addr4);
    print_free_lists();

    printf("Freeing block at address: 0x%llx\n", addr5);
    buddy_free_(addr5);
    print_free_lists();

    // Stress test with maximum allocations and coalescing
    printf("Stress testing allocations and frees...\n");
    unsigned long long addresses[100];
    int allocation_count = 0;

    for (int size = 8; size <= 64; size *= 2) {
        for (int i = 0; i < 10; i++) {
            unsigned long long addr = buddy_alloc_(size);
            if (addr != (unsigned long long)-1) {
                addresses[allocation_count++] = addr;
                printf("Allocated block of size %d at address: 0x%llx\n", size, addr);
            }
        }
    }

    print_free_lists();

    printf("Freeing all blocks in reverse order...\n");
    for (int i = allocation_count - 1; i >= 0; i--) {
        printf("Freeing block at address: 0x%llx\n", addresses[i]);
        buddy_free_(addresses[i]);
    }

    print_free_lists();

    printf("Intensive test for buddy_free_ completed.\n");
}
