// slab.c
#include "slub.h"
#include "buddy2.h"
#include "defs.h"

const uint32 kmem_cache_objsize[] = {8, 16, 32, 64, 128, 256, 512, 1024, 2048};
const char *kmem_cache_name[] = {
    "kmem_cache-8   ", "kmem_cache-16  ", "kmem_cache-32  ",
    "kmem_cache-64  ", "kmem_cache-128 ", "kmem_cache-256 ",
    "kmem_cache-512 ", "kmem_cache-1024", "kmem_cache-2048"};

struct kmem_cache *slub_allocator[SLUB_NUM] = {};

struct page pages[PGREF_MAX_ENTRIES];
struct spinlock pages_lock;
struct cache_area cache_region;

uint32 cache_tid = 0; //global cache transaction id


void page_init() { 
    for(uint32 i = 0; i < PA2PGREF_ID(bdsys.base_addr); ++i){
        pages[i].refcount = 0;
        pages[i].freelist = 0;
        pages[i].flags = PAGE_RESERVE;
        pages[i].slab = 0;
        pages[i].header = 0;
        pages[i].next = 0;
        pages[i].active_obj_count = 0;
    }
    for(uint32 i = PA2PGREF_ID(bdsys.base_addr); i < PGREF_MAX_ENTRIES; ++i){
        pages[i].refcount = 0;
        pages[i].freelist = 0;
        pages[i].flags = PAGE_FREE;
        pages[i].slab = 0;
        pages[i].header = 0;
        pages[i].next = 0;
        pages[i].active_obj_count = 0;
    }
}

//调用本函数前请先为objsize做8字节对齐处理，否则指针会无法正确初始化
void *init_object_list(void *addr, uint32 objsize, uint32 size) {
    uint64 end = (uint64)addr + size;
    uint64 this_obj = (uint64)addr;
    uint64 next_obj = (uint64)addr + objsize;

    while (next_obj < end) {
        *(void **)(this_obj) = (void *)next_obj;
        this_obj += objsize;
        next_obj += objsize;
    }

    if (next_obj == end) {
        *(void **)(this_obj) = 0;
        return (void *)this_obj;
    } else {
        *(void **)(this_obj - objsize) = 0; //超出则舍去
        return (void *)(this_obj - objsize);
    }
}

void kmem_cache_init() {
    void *kmem_freelist = (void*)(buddy_alloc_(KMEM_PAGE_NUM)); //为kmem_cache分配一段8页的空间，加以链表的组织结构。我们要求kmem_cache 64位对齐。

    if (kmem_freelist == 0)
        panic("kmem_cache_init");

    for(uint32 i = PA2PGREF_ID((uint64)kmem_freelist); i < PA2PGREF_ID((uint64)kmem_freelist) + KMEM_PAGE_NUM; ++i){ //对这里的页面做些简单的初始化，其实不做也行
        pages[i].refcount = 0;
        pages[i].freelist = 0;
        pages[i].flags = PAGE_RESERVE;
        pages[i].slab = 0;
        pages[i].header = 0;
        pages[i].next = 0;
        pages[i].active_obj_count = 0;
    }
    cache_region.base = kmem_freelist; //freelist由cache_region统一管理。接下来我们按8字节对齐初始化free list:

    init_object_list(kmem_freelist, ALIGN_SIZE(sizeof(struct kmem_cache), 8), KMEM_PAGE_NUM << 12);
    cache_region.freelist = (void **)kmem_freelist;
}




struct kmem_cache *alloc_kmem_cache() {
  struct kmem_cache *p;

  if (cache_region.freelist == 0) return 0;

  p = (struct kmem_cache *)(cache_region.freelist);
  cache_region.freelist = *(void **)(cache_region.freelist);
  memset(p, 0, sizeof(struct kmem_cache));

  return p;
}

static void inline free_kmem_cache(struct kmem_cache *cache) {
  if (cache_region.freelist == 0)
    cache_region.freelist = (void *)cache;
  else {
    *((void**)cache) = cache_region.freelist;
    cache_region.freelist = (void *)cache;
  } 
}

void *cache_create(const char *name, uint32 size, uint32 aligns, int flags, void *func(void *)) { //设置cache元信息
    struct kmem_cache *s = 0;
    s = alloc_kmem_cache();
    if (s == 0) {
        panic("cache_create");
    }

    /* init kmem_cache */
    s->name = name;
    s->ctor = (void *)func;
    s->refcount = 1;

    s->min_partial = 4;
    s->size = ALIGN_SIZE(size, aligns);
    s->size = s->size < 8 ? 8 : s->size;  //对象+padding后的大小
    s->object_size = size;                //对象的实际大小
    s->offset = 0;                        //自由指针的位置
    s->inuse = 0;
    s->align = aligns;                    //对齐要求

    s->nr_page_per_slub = PAGE_NUM_PER_SLUB(s->size);
    s->nr_partial = 0;
    s->tid = cache_tid++;
    return s;
}

//首先从cpu 本地缓存池分配，如果freelist不存在，
//就会转向per cpu partial分配，如果per cpu partial也没有可用对象，
//继续查看per node partial，如果很不幸也不没有可用对象的话，
//就只能从伙伴系统分配一个slab了，并挂入per cpu freelist。
//此时调用本函数:
void *cache_alloc_pages(struct kmem_cache *cache) {
    if(cache == 0)
        return 0;
    void *slab = (void*)buddy_alloc_(cache->nr_page_per_slub);
    if(slab == 0) {
        return 0;
    }
    memset(slab, 0, (cache->nr_page_per_slub) << 12);
    int start = PA2PGREF_ID((uint64)slab);
    for(int i = start; i < start + cache->nr_page_per_slub; i++){
        pages[i].refcount = 0;
        pages[i].slab = cache;
        pages[i].flags = PAGE_SLUB;
        pages[i].header = &pages[start];
        pages[i].freelist = 0; 
        pages[i].next = 0;
    }//晚些我们会设置header页的各个字段，如next, freelist

    pages[start].next = cache->partial_list; //can be null
    cache->partial_list = &pages[start];

    init_object_list(slab, cache->size, ((cache->nr_page_per_slub) << 12));
    cache->freelist = (void**)slab;
    cache->page = &pages[start];
    cache->nr_partial++;
    pages[start].freelist = (void*) slab;
    return slab;
}

struct kmem_cache *kmem_cache_create(const char *name, uint32 size, uint32 aligns, int flags, void *func(void *)) {
    struct kmem_cache *s = 0;
    s = cache_create(name, size, aligns, flags, func);
    if (cache_alloc_pages(s) == 0) {
        free_kmem_cache(s);
        printf("cache fails to create");
        return 0;
    } else {
        return s;
    }
}


//怎样才能destroy一个cache？
//首先对象得已经完全归还，不然这时把cache摧毁会带来一致性问题

int kmem_cache_destroy(struct kmem_cache *cache) {
    if(cache == 0) {
        printf("cache is null");
        return 0;
    }
    //不妨开始检查是否已经归还完毕，如果是，那么释放其所有资源。
    for(struct page *p = cache->partial_list; p != 0; p = p->next) {
        if(p->active_obj_count > 0)
            return -1;
    }
    for(struct page *p = cache->partial_list; p != 0; p = p->next) {
        buddy_free_(((p - &pages[0]) << 12) + KERNBASE);
        for(uint32 i = p-&pages[0]; i < p-&pages[0] + cache->nr_page_per_slub; ++i){
            pages[i].refcount = 0;
            pages[i].freelist = 0;
            pages[i].flags = PAGE_FREE;
            pages[i].slab = 0;
            pages[i].header = 0;
            pages[i].next = 0;
            pages[i].active_obj_count = 0;
        }
    }
    return 0;
}

//当我们使用某个partial list中的slab时，我们也不会把它从list里拿出来。
//当它为空后，cache_alloc会往partial list头插新的slab。
//在free的时候，会直接填入正确的位置，如果条件满足，就把它从partial list中拿出并还给buddy system
void *kmem_cache_alloc(struct kmem_cache *cache) {
    if(cache == 0)
        return 0;
    void *object = 0;
    if (cache->freelist == 0) {
        for(struct page *p = cache->partial_list; p->next != 0; p = p->next){
            if(p->active_obj_count) {
                cache->page = p;
                cache->freelist = (void**)p->freelist;
                break;
            }
        }
        if(cache->freelist == 0) {
            cache_alloc_pages(cache);

        }
        if(cache->freelist == 0) {
            return 0;
        }
    }

    object = (void*)cache->freelist;
    cache->freelist = *(cache->freelist);
    cache->page->freelist = cache->freelist;

    pages[PA2PGREF_ID((uint64)object)].header->active_obj_count++;  
    if (cache->ctor != 0)
        cache->ctor(object);
    else {
        memset(object, 0, cache->size);
    }
    
    return object;
}

void test_kmem_cache_full(struct kmem_cache *cache){
    int i = 0;
    for(struct page *p = cache->partial_list; p != 0; p = p->next) {
        i++;
        if(p->active_obj_count != 0)
            panic("kmem_cache_full test failed");
        
    }
    if(i != cache->nr_partial)
        panic("kmem_cache_full test failed");
}

//这个测试必须只能在full test之后
void test_kmem_cache_nr_partial(struct kmem_cache *cache){
    if(cache->nr_partial > cache->min_partial)
        panic("nr partial test failed");
}

//这个测试必须只能在full test之后
void test_kmem_cache_free_list(struct kmem_cache *cache){
    for(struct page *p = cache->partial_list; p != 0; p = p->next) {
        int i = 0;
        for(void* ptr = p->freelist; ptr != 0; ptr = *((void**)ptr)){
            i++;
        }
        if(i != (cache->nr_page_per_slub << 12) / cache->object_size) {
            printf("count: %d, actual: %ld", i, (cache->nr_page_per_slub << 12) / cache->object_size);
            panic("free list test fail");
        }
    }
}


void kmem_cache_free(void *obj) {
    if(obj == 0)
        return;
    struct page *header = pages[PA2PGREF_ID((uint64)obj)].header;
    struct kmem_cache *cache = header->slab;

    *((void**)obj) = header->freelist;
    header->freelist = obj;
    if(header == cache->page)
        cache->freelist = obj;
    header->active_obj_count--;
    
    if(header->active_obj_count == 0 && cache->min_partial < cache->nr_partial && header != cache->page){
        //删除partial list中的header
        struct page *ptr = cache->partial_list;
        //由于不可能为空
        for(; ptr->next != header; ptr = ptr->next){
            if(ptr == 0)
                panic("kmem_cache_free");
        }
        ptr->next = header->next;

        for(uint32 i = header-&pages[0]; i < header-&pages[0] + cache->nr_page_per_slub; ++i){
            pages[i].refcount = 0;
            pages[i].freelist = 0;
            pages[i].flags = PAGE_FREE;
            pages[i].slab = 0;
            pages[i].header = 0;
            pages[i].next = 0;
            pages[i].active_obj_count = 0;
        }
        buddy_free_(((header - &pages[0]) << 12) + KERNBASE);

        cache->nr_partial--;
    }
}

void slub_allocator_init() {
  page_init();
  kmem_cache_init();
  for(int i = 0; i < SLUB_NUM; i++){
    slub_allocator[i] = kmem_cache_create(kmem_cache_name[i], kmem_cache_objsize[i], 8, 0, 0);
  }
  return;
}


void *kmalloc_(uint32 size) {
  int objindex;
  void *p = 0;
  if (size == 0) return 0;

  for (objindex = 0; objindex < SLUB_NUM; objindex++) {
    if(size <= kmem_cache_objsize[objindex]) {
      p = kmem_cache_alloc(slub_allocator[objindex]);
      return p;
    }
  }

  // size 若不在 kmem_cache_objsize 范围之内，则使用 buddy system 来分配内存
  if (objindex >= SLUB_NUM) {
    uint32 s = 1024;
    while(s < size)
        s = s << 1;

    p = (void*)buddy_alloc_((size - 1) / PGSIZE + 1);
    
  }

  return p;
}

void kfree_(const void *addr) {
  struct page *page;

  
  
  if (addr == 0) return;

  page = &pages[PA2PGREF_ID((uint64)addr)];
  if (page->flags == PAGE_BUDDY) {
    if(((uint64)addr & 0xfff) != 0)
        panic("not aligned in kfree_");

    buddy_free_((uint64)addr);


  } else if (page->flags == PAGE_SLUB) {

    kmem_cache_free((void *)addr);
  } else {
    panic("kfree1");
  }

  return;
}

#define TEST_ROUNDS 300
#define MAX_ALLOC_SIZE 2048

void simple_allocation_test() {
    printf("Running simple allocation test...\n");

    void *ptr1 = kmalloc(64);
    void *ptr2 = kmalloc(128);

    if (!ptr1 || !ptr2) {
        panic("Test failed: kmalloc returned null for valid allocation size.\n");
    }

    printf("Simple allocation successful. Freeing memory...\n");
    kfree(ptr1);
    kfree(ptr2);
    for(int i = 0; i < SLUB_NUM; ++i) {
        test_kmem_cache_full(slub_allocator[i]);
        test_kmem_cache_nr_partial(slub_allocator[i]);
        test_kmem_cache_free_list(slub_allocator[i]);
    }
    printf("Simple allocation test passed.\n");
}

void stress_test() {
    printf("Running stress test with %d allocations...\n", TEST_ROUNDS);

    void *ptrs[TEST_ROUNDS];
    for (int i = 0; i < TEST_ROUNDS; i++) {
        int size = (i % MAX_ALLOC_SIZE) + 1; // Vary sizes from 1 to MAX_ALLOC_SIZE
        ptrs[i] = kmalloc(size);

        if (!ptrs[i]) {
            printf("Test failed: kmalloc failed on round %d for size %d.\n", i, size);
            panic("test fail");
        }
    }

    printf("Stress allocation successful. Freeing memory...\n");

    for (int i = 0; i < TEST_ROUNDS; i++) {
        kfree(ptrs[i]);
    }
    for(int i = 0; i < SLUB_NUM; ++i) {
        test_kmem_cache_full(slub_allocator[i]);
        test_kmem_cache_nr_partial(slub_allocator[i]);
        test_kmem_cache_free_list(slub_allocator[i]);
    }
    printf("Stress test passed.\n");
}
