#include "vma.h"
#include "param.h"
#include "memlayout.h"
#include "spinlock.h"
#include "proc.h"
#include "defs.h"

void init_mm_struct(struct mm_struct *mm) {
    mm->mmap = 0;
    mm->lower_bound = PGROUNDDOWN(MMAPEND);
}


struct vm_area_struct *find_vma(struct mm_struct *mm, uint64 addr){
    struct vm_area_struct *ptr;

    if(mm == 0 || (ptr = mm->mmap) == 0){
        return 0;
    }
    while(ptr != 0){
        if(ptr->vm_start <= addr && ptr->vm_end > addr){
            return ptr;
        }
        ptr = ptr->vm_next;
    }
    return 0;
}

//addr is the hint for the start address of the VMA. If vm_type is not VMA_FILE, then we will use it.
//len is the length of mapping, has to be page-aligned.
//vm_sizesz is the file size after the offset, can be not page-aligned.
//vm_pgoff is the offset within the file at the vm_start, can be not page-aligned.
uint64 do_mmap(struct mm_struct *mm, uint64 addr, uint64 len, uint64 vm_pgoff, uint64 
    vm_filesz, int flags, struct file *f, int prot, int vma_type){
    if(len == 0)
        return -1;
    if((addr & 0xfff) != 0)
        return -1;
    struct vm_area_struct *ptr = (struct vm_area_struct*) kmalloc(sizeof(struct vm_area_struct));
    if(ptr == 0)
        return -1;

    uint64 end_addr = 0, start_addr = 0;
    if(vma_type == VMA_FILE) {
        // printf("mm->lower bound : %lx", mm->lower_bound);
        end_addr = mm->lower_bound;
        start_addr = end_addr - PGROUNDUP(len);
        mm->lower_bound = start_addr;
    } else {
        start_addr = addr;
        end_addr = addr += PGROUNDUP(len);
    }

    if(end_addr == 0)
        return -1;

    //addr parameter should be ingnored
    ptr->vm_end = end_addr;
    ptr->vm_start = PGROUNDDOWN(start_addr);
    
    ptr->vm_flags = flags;
    ptr->vm_pgoff = vm_pgoff;

    ptr->vm_filesz = vm_filesz;

    ptr->vm_file = f;
    ptr->vm_prot = prot;
    
    ptr->vma_type = vma_type;

    //insert the VMA into the list
    ptr->vm_next = mm->mmap;
    if(mm->mmap != 0)
        mm->mmap->vm_prev = ptr;
    ptr->vm_prev = 0;
    mm->mmap = ptr;

    return ptr->vm_start;
}

int test_vma_all_unmapped(struct vm_area_struct *vma) {
    if((vma->vm_start & 0xfff) != 0)
        panic("test_vma_all_unmapped");
    if((vma->vm_end & 0xfff) != 0)
        panic("test_vma_all_unmapped");
    for(uint64 va = vma->vm_start; va < vma->vm_end; va += PGSIZE) {
        pte_t *pte = walk(myproc()->pagetable, va, 0);
        if((*pte & PTE_V) != 0)
            return 0;
    }
    return 1;
}

int unmap_all_vma(struct mm_struct *mm) {
    struct vm_area_struct *vma = mm->mmap;

    while (vma) {
        struct vm_area_struct *next = vma->vm_next;
         unmap_mmap_area(myproc()->pagetable, vma->vm_start, 
                         (vma->vm_end - vma->vm_start) >> 12, vma);
        if (vma->vma_type == VMA_FILE && vma->vm_file) {
            fileclose(vma->vm_file);
        }
        kfree(vma);
        vma = next;
    }
    init_mm_struct(mm);
    return 0;
}

struct vm_area_struct *copy_vma_list(struct vm_area_struct *parent_vma) {
    struct vm_area_struct *new_vma_head = 0;
    struct vm_area_struct *new_vma_tail = 0;

    while (parent_vma) {
        struct vm_area_struct *new_vma = (struct vm_area_struct *)kmalloc(sizeof(*parent_vma));
        if (!new_vma) {
            panic("copy_vma_list");
        }

        new_vma->vm_start = parent_vma->vm_start;
        new_vma->vm_end = parent_vma->vm_end;
        new_vma->vm_flags = parent_vma->vm_flags;
        new_vma->vm_prot = parent_vma->vm_prot;
        new_vma->vm_pgoff = parent_vma->vm_pgoff;
        new_vma->vm_filesz = parent_vma->vm_filesz;
        new_vma->vma_type = parent_vma->vma_type;


        if (parent_vma->vm_file) {
            new_vma->vm_file = filedup(parent_vma->vm_file); 
        } else {
            new_vma->vm_file = 0;
        }

        new_vma->vm_next = 0;
        new_vma->vm_prev = new_vma_tail;

        if (new_vma_tail) {
            new_vma_tail->vm_next = new_vma; 
        } else {
            new_vma_head = new_vma; 
        }
        new_vma_tail = new_vma; 

        parent_vma = parent_vma->vm_next;
    }

    return new_vma_head;
}

void print_vma(struct vm_area_struct *vma) {
    if (vma == 0) {
        printf("VMA is NULL\n");
        return;
    }

    printf("VM Area Struct:\n");
    printf("  Start Address: 0x%lx\n", vma->vm_start);
    printf("  End Address: 0x%lx\n", vma->vm_end);
    printf("  Flags: %d\n", vma->vm_flags);
    printf("  Protection: %d\n", vma->vm_prot);
    printf("  File: %p\n", vma->vm_file);
    printf("  Page Offset: %lx\n", vma->vm_pgoff);
    printf("  File Size: %lx\n", vma->vm_filesz);
    printf("  VMA Type: 0x%x (", vma->vma_type);
    if (vma->vma_type == VMA_FILE) printf("VMA_FILE");
    if (vma->vma_type == VMA_HEAP) printf("VMA_HEAP");
    if (vma->vma_type == VMA_TEXT) printf("VMA_TEXT");
    if (vma->vma_type == VMA_STACK) printf("VMA_STACK");
    if (vma->vma_type == VMA_ANON) printf("VMA_ANON");
    if (vma->vma_type == VMA_DATA) printf("VMA_DATA");
    printf(")\n");
}

void print_vma_list(struct vm_area_struct *head) {
    struct vm_area_struct *current = head;

    if (current == 0) {
        printf("VMA list is empty\n");
        return;
    }

    while (current) {
        print_vma(current);
        current = current->vm_next; // Move to the next VMA
    }
}